"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Transmit = exports.TRANSMIT_DEFAULT_OPTIONS = void 0;
const busboy_1 = __importDefault(require("busboy"));
const stream_1 = __importDefault(require("stream"));
const events_1 = require("events");
const util_1 = require("util");
const transmit_managers_1 = require("./transmit.managers");
const transmit_exceptions_1 = require("./transmit.exceptions");
const counter_util_1 = require("./utils/counter.util");
const hash_calculator_util_1 = require("./utils/hash-calculator.util");
const stream_meter_util_1 = require("./utils/stream-meter.util");
const generate_random_string_util_1 = require("./utils/generate-random-string.util");
const pipeline = util_1.promisify(stream_1.default.pipeline);
exports.TRANSMIT_DEFAULT_OPTIONS = {
    hashAlgorithm: "sha256",
    manager: new transmit_managers_1.DiskManager(),
    minFields: 0,
    maxFields: 100,
    minFiles: 1,
    maxFiles: 1,
    maxFieldNameSize: 100,
    maxFieldValueSize: 1 * 1024 * 1024,
    minFileSize: 0,
    maxFileSize: 100 * 1024 * 1024,
    maxHeaderPairs: Infinity,
    maxParts: Infinity,
    transformers: [],
    truncateFieldNames: false,
    truncateFieldValues: false
};
class Transmit extends events_1.EventEmitter {
    constructor(options = {}) {
        super();
        this.fields = [];
        this.files = [];
        this.incoming = [];
        this.pendingFiles = new counter_util_1.Counter();
        this.isAborted = false;
        this.isFinished = false;
        this.options = Object.assign(Object.assign({}, exports.TRANSMIT_DEFAULT_OPTIONS), options);
    }
    /**
     * Whether the upload has been aborted.
     */
    get aborted() {
        return this.isAborted;
    }
    /**
     * Whether the upload has been finished.
     */
    get finished() {
        return this.isFinished;
    }
    /**
     * The transmit manager used for this instance.
     */
    get manager() {
        return this.options.manager;
    }
    /**
     * Abort the current upload. Emits "aborted" and "done" events.
     */
    abort(error) {
        if (this.aborted || this.finished) {
            return;
        }
        this.isAborted = true;
        this.pendingFiles.whenItEqualsTo(0, () => {
            this.emit("aborted", error);
            this.emit("done");
        });
    }
    /**
     * Delete uploaded files for this request. Useful for cleanup after aborting the upload.
     */
    removeUploadedFiles() {
        return __awaiter(this, void 0, void 0, function* () {
            const tasks = yield Promise.allSettled(this.incoming.map((file) => this.manager.removeFile(file)));
            for (const task of tasks) {
                if (task.status === "rejected") {
                    throw task.reason;
                }
            }
        });
    }
    /**
     * Parse an incoming request with form data.
     */
    parse(req) {
        const busboy = this.createBusboy({
            headers: req.headers,
            limits: {
                fieldNameSize: this.options.maxFieldNameSize,
                fieldSize: this.options.maxFieldValueSize,
                fields: this.options.maxFields,
                fileSize: this.options.maxFileSize,
                files: this.options.maxFiles,
                headerPairs: this.options.maxHeaderPairs,
                parts: this.options.maxParts
            }
        });
        req.on("aborted", () => {
            busboy.end();
            this.abort();
        });
        req.pipe(busboy);
        return busboy;
    }
    /**
     * Parse an incoming request with form data and return its fields and files.
     * If the request is aborted, all uploaded files will be deleted.
     */
    parseAsync(req) {
        const busboy = this.parse(req);
        return new Promise((resolve, reject) => {
            this.once("aborted", (error) => __awaiter(this, void 0, void 0, function* () {
                yield this.removeUploadedFiles();
                if (error) {
                    return reject(error);
                }
                resolve({
                    fields: [],
                    files: []
                });
            }));
            this.once("done", () => {
                req.unpipe(busboy);
                // Drain the stream to make sure req emits "end" event
                // https://nodejs.org/api/stream.html#stream_event_end
                // https://nodejs.org/api/stream.html#stream_readable_resume
                req.resume();
                busboy.removeAllListeners();
            });
            this.once("finished", (files, fields) => {
                resolve({
                    fields,
                    files
                });
            });
        });
    }
    createBusboy(options) {
        try {
            return new busboy_1.default(options)
                .on("field", (...args) => this.handleIncomingField(...args))
                .on("file", (...args) => {
                this.pendingFiles.increment();
                this.handleIncomingFile(...args)
                    .finally(() => this.pendingFiles.decrement())
                    .then(() => this.finish())
                    .catch((error) => this.abort(error));
            })
                .once("error", (error) => this.abort(error))
                .once("fieldsLimit", () => this.abort(new transmit_exceptions_1.TooManyFieldsException()))
                .once("filesLimit", () => this.abort(new transmit_exceptions_1.TooManyFilesException()))
                .once("finish", () => this.finish())
                .once("partsLimit", () => this.abort(new transmit_exceptions_1.TooManyPartsException()));
        }
        catch (_a) {
            throw new transmit_exceptions_1.UnsupportedContentTypeException();
        }
    }
    finish() {
        if (this.aborted || this.finished || !this.pendingFiles.is(0)) {
            return;
        }
        if (this.fields.length < this.options.minFields) {
            return this.abort(new transmit_exceptions_1.NotEnoughFieldsException());
        }
        if (this.files.length < this.options.minFiles) {
            return this.abort(new transmit_exceptions_1.NotEnoughFilesException());
        }
        // this.options.maxFiles is used in busboy config so no need to check that
        this.isFinished = true;
        this.emit("finished", this.files, this.fields);
        this.emit("done");
    }
    handleIncomingField(fieldName, fieldValue, fieldNameTruncated, fieldValueTruncated, encoding, mimetype) {
        // workaround for busboy bug, see https://github.com/mscdex/busboy/pull/59
        if (fieldName.length > this.options.maxFieldNameSize) {
            fieldName = fieldName.substr(0, this.options.maxFieldNameSize);
            fieldNameTruncated = true;
        }
        if (fieldNameTruncated && !this.options.truncateFieldNames) {
            return this.abort(new transmit_exceptions_1.FieldNameTooLargeException());
        }
        if (fieldValueTruncated && !this.options.truncateFieldValues) {
            return this.abort(new transmit_exceptions_1.FieldValueTooLargeException());
        }
        this.fields.push({
            encoding,
            mimetype,
            name: fieldName,
            value: String(fieldValue)
        });
    }
    handleIncomingFile(field, readable, filename, encoding, mimetype) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.aborted || this.finished) {
                return readable.resume();
            }
            // Ignore files with empty filenames
            if (!filename || filename.length <= 0) {
                return readable.resume();
            }
            if (this.options.field && field !== this.options.field) {
                return readable.resume();
            }
            const id = yield generate_random_string_util_1.generateRandomString();
            const incomingFile = {
                encoding,
                field,
                id,
                mimetype,
                name: filename
            };
            const accept = this.options.filter ? yield this.options.filter(incomingFile) : true;
            if (!accept) {
                return readable.resume();
            }
            this.incoming.push(incomingFile);
            // Create PR and add truncated property to readable stream
            // https://github.com/DefinitelyTyped/DefinitelyTyped/blob/master/types/busboy/index.d.ts#L41
            yield this.writeFile(incomingFile, readable);
        });
    }
    writeFile(file, readable) {
        return __awaiter(this, void 0, void 0, function* () {
            const writable = yield this.manager.handleFile(file);
            const hash = new hash_calculator_util_1.HashCalculator(this.options.hashAlgorithm);
            const sizeBeforeTransformers = new stream_meter_util_1.StreamMeter();
            const sizeAfterTransformers = new stream_meter_util_1.StreamMeter();
            const transformers = this.options.transformers.map((transform) => transform(file));
            yield pipeline(readable, sizeBeforeTransformers, // Capture the file size before the file is modified
            ...transformers, sizeAfterTransformers, // Capture the file size after the file has been modified
            hash, writable);
            if (readable.truncated) {
                throw new transmit_exceptions_1.FileTooLargeException();
            }
            if (sizeBeforeTransformers.size < this.options.minFileSize) {
                throw new transmit_exceptions_1.FileTooSmallException();
            }
            this.files.push(Object.assign(Object.assign({}, file), { hash: hash.digest, size: sizeAfterTransformers.size }));
        });
    }
}
exports.Transmit = Transmit;
